#include "../include/instruction_handler.h"
#include "../include/logger.h"

InstructionHandler::InstructionHandler() {
    this->handlers[0x0000] = handle0000;
    this->handlers[0x1000] = handle1000;
    this->handlers[0x2000] = handle2000;
    this->handlers[0x3000] = handle3000;
    this->handlers[0x4000] = handle4000;
    this->handlers[0x5000] = handle5000;
    this->handlers[0x6000] = handle6000;
    this->handlers[0x7000] = handle7000;
}

/**
 * Attempts to handle the Chip8 instruction by the given instruction number.
 * If the handler doesn't exist for the instruction, an error is logged, but the program resumes.
 *
 * @param instructionNumber the 16-bit value to determine the instruction
 * @param chip8 the Chip8 instance
 * @param instruction the Instruction instance
 */
void InstructionHandler::handleInstruction(unsigned short instructionNumber, Chip8& chip8, Instruction instruction) {
    if (!this->handlers.contains(instructionNumber)) {
        Logger::error("Unregistered instruction", instructionNumber);
        return;
    }

    this->handlers[instructionNumber](chip8, instruction);
}

/**
 * Handles the 0x0000 Chip-8 instruction.
 * This handles 2 other instructions:
 * - 0x00E0 which is used to clear the screen.
 * - 0x00EE which is used to return from a subroutine.
 *
 * In older interpreters, 0x0NNN was also used to jump to a machine code at NNN, but this shouldn't be implemented.
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle0000(Chip8& chip8, Instruction instruction) {
    switch (instruction.kk()) {
        case 0x00E0:
            chip8.getDisplay().clear();
            break;
        case 0x00EE:
            chip8.setProgramCounter(chip8.popFromStack());
            break;
    }
}

/**
 * Handles the 0x1NNN Chip8 instruction which is used to jump to location NNN.
 * This will set the program counter to the NNN value.
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle1000(Chip8& chip8, Instruction instruction) {
    chip8.setProgramCounter(instruction.nnn());
}

/**
 * Handles the 0x2NNN Chip8 instruction which is used to call a subroutine at NNN.
 * This will increments the stack pointer and puts the program counter on top of the stack.
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle2000(Chip8& chip8, Instruction instruction) {
    chip8.pushToStack(chip8.getProgramCounter());
    chip8.setProgramCounter(instruction.nnn());
}

/**
 * Handles the 0x3XKK Chip8 instruction which is used to skip the next instruction if V[x] == kk.
 * This will compare V[x] to kk and increases the program counter by 2 if they are equal.
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle3000(Chip8& chip8, Instruction instruction) {
    if (chip8.getRegister(instruction.x()) == instruction.kk()) {
        chip8.setProgramCounter(chip8.getProgramCounter() + 2);
    }
}

/**
 * Handles the 0x4XKK Chip8 instruction which is used to skip the next instruction if V[x] != kk.
 * This will compare V[x] to kk and increases the program counter by 2 if they are not equal.
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle4000(Chip8& chip8, Instruction instruction) {
    if (chip8.getRegister(instruction.x()) != instruction.kk()) {
        chip8.setProgramCounter(chip8.getProgramCounter() + 2);
    }
}

/**
 * Handles the 0x5XY0 Chip8 instruction which is used to skip the next instruction if V[x] == V[y].
 * This will compare V[x] to V[y] and increases the program counter by 2 if they are equal.
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle5000(Chip8& chip8, Instruction instruction) {
    if (chip8.getRegister(instruction.x()) == chip8.getRegister(instruction.y())) {
        chip8.setProgramCounter(chip8.getProgramCounter() + 2);
    }
}

/**
 * Handles the 0x6XKK Chip8 instruction which is used to put value KK in register V[x].
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle6000(Chip8& chip8, Instruction instruction) {
    chip8.setRegister(instruction.x(), instruction.kk());
}

/**
 * Handles the 0x7XKK Chip8 instruction which is used to add the value KK to register V[x] and set the result in register V[x].
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle7000(Chip8& chip8, Instruction instruction) {
    chip8.setRegister(instruction.x(), chip8.getRegister(instruction.x()) + instruction.kk());
}

/**
 * Handles the 0x8XY* opcodes (0x8XY0 to 0x8XY7 + 0x8XYE).
 * The following opcodes are handled:
 *  - 0x8XY0 - Stores the value of register V[y] into register V[x]
 *  - 0x8XY1 - Stores the bitwise OR result of register V[x] and register V[y] into register V[x]
 *  - 0x8XY2 - Stores the bitwise AND result of register V[x] and register V[y] into register V[x]
 *  - 0x8XY3 - Stores the bitwise XOR result of register V[x] and register V[y] into register V[x]
 *  - 0x8XY4 - Sum of register V[y] and register V[x] is checked, if the result is greater than 8 bits, register V[F] is
 *             set to 1, otherwise 0. Only lowest 8 bits of the result are kept and stored into register V[x]
 *  - 0x8XY5 - If register V[x] is higher than register V[y], set register V[F] to 1, otherwise 0.
 *             Store the result of register V[x] - register V[y] into register V[x].
 *  - 0x8XY6 - If the least significant bit of register V[x] is 1, set register V[F] to 1, otherwise 0.
 *             Divide register V[x] by 2 (shift right with 1) and store the result in register V[x].
 *  - 0x8XY7 - If register V[y] is higher than register V[x], set register V[F] to 1, otherwise 0.
 *             Store the result of register V[y] - register V[x] into register V[x].
 *  - 0x8XYE - If the most significant bit of register V[x] is 1, set register V[F] to 1, otherwise 0.
 *             Multiply register V[x] by 2 (shift left with 1) and store the result in register V[x].
 *
 * @param chip8 The Chip8 instance
 * @param instruction The Instruction instance
 */
void InstructionHandler::handle8000(Chip8& chip8, Instruction instruction) {
    switch (instruction.n()) {
        case 0: // 0x8XY0
            chip8.setRegister(instruction.x(), chip8.getRegister(instruction.y()));
            break;
        case 1: // 0x8XY1
            chip8.setRegister(instruction.x(), chip8.getRegister(instruction.x()) | chip8.getRegister(instruction.y()));
            break;
        case 2: // 0x8XY2
            chip8.setRegister(instruction.x(), chip8.getRegister(instruction.x()) & chip8.getRegister(instruction.y()));
            break;
        case 3: // 0x8XY3
            chip8.setRegister(instruction.x(), chip8.getRegister(instruction.x()) ^ chip8.getRegister(instruction.y()));
            break;
        case 4: // 0x8XY4
        {
            unsigned short sum = chip8.getRegister(instruction.y()) + chip8.getRegister(instruction.x());
            chip8.setRegister(0xF, sum > 0xFF ? 1 : 0);
            chip8.setRegister(instruction.x(), sum & 0xFF);
            break;
        }
        case 5: // 0x8XY5
        {
            unsigned short regX = chip8.getRegister(instruction.x());
            unsigned short regY = chip8.getRegister(instruction.y());

            chip8.setRegister(0xF, regX > regY ? 1 : 0);
            chip8.setRegister(instruction.x(), (regX - regY) & 0xFF);
            break;
        }
        case 6: // 0x8XY6
        {
            unsigned short regX = chip8.getRegister(instruction.x());
            chip8.setRegister(0xF, regX & 0x1);
            chip8.setRegister(instruction.x(), regX >> 1);
            break;
        }
        case 7: // 0x8XY7
        {
            unsigned short regX = chip8.getRegister(instruction.x());
            unsigned short regY = chip8.getRegister(instruction.y());

            chip8.setRegister(0xF, regY > regX ? 1 : 0);
            chip8.setRegister(instruction.x(), (regY - regX) & 0xFF);
            break;
        }
        case 0xE: // 0x8XYE
        {
            unsigned short regX = chip8.getRegister(instruction.x());
            chip8.setRegister(0xF, ((regX >> 7) & 0x1) == 1);
            chip8.setRegister(instruction.x(), (regX << 1) & 0xFF);
            break;
        }
    }
}
